From f8c21dce473e7835c2435ff9cca92ef9e6c573ad Mon Sep 17 00:00:00 2001
From: "Robert A. Burks" <rburksdev@gmail.com>
Date: Thu, 14 Mar 2024 21:56:22 -0400
Subject: [PATCH 16/31] Fixes for two variable watcher bugs (BUG#00004)
 (BUG#00005)

* src/data.c (set_internal): Will now determine case to notify 'where'.
(set_default_internal): Will now determine case to notify 'where'
(notify_variable_watchers): Calling functions will now determine 'where'.
Variable watchers will now be properly notified of the 'unlet'
(BUG#00004) and "where" 'makunbound' occurred (BUG#00005)
for buffer local variables.
---
 src/data.c | 46 ++++++++++++++++++++++++++++++++++------------
 1 file changed, 34 insertions(+), 12 deletions(-)

diff --git a/src/data.c b/src/data.c
index fe9d097ec30..9fd481caa19 100644
--- a/src/data.c
+++ b/src/data.c
@@ -1672,10 +1672,22 @@ set_internal (Lisp_Object symbol, Lisp_Object newval, Lisp_Object where,
 	/* If watched notify unless thread switching.  */
 	if (sym->u.s.trapped_write == SYMBOL_TRAPPED_WRITE
 	    && bindflag != SET_INTERNAL_THREAD_SWITCH)
-	  notify_variable_watchers (symbol, voide? Qnil : newval,
-				    (bindflag == SET_INTERNAL_BIND? Qlet :
-				     bindflag == SET_INTERNAL_UNBIND? Qunlet :
-				     voide? Qmakunbound : Qset), where);
+	  {
+	    /* Only set 'where' if there exists a local value
+	       or if we are setting a local_if_set while not
+	       shadowing default in a let.  */
+	    if (NILP (where)
+		&& (!NILP (Flocal_variable_p (symbol, Fcurrent_buffer ()))
+		    || (bindflag == SET_INTERNAL_SET
+			&& blv->local_if_set
+		        && !let_shadows_buffer_binding_p (sym))))
+	      XSETBUFFER (where, current_buffer);
+
+	    notify_variable_watchers (symbol, voide? Qnil : newval,
+				      (bindflag == SET_INTERNAL_BIND? Qlet :
+				       bindflag == SET_INTERNAL_UNBIND? Qunlet :
+				       voide? Qmakunbound : Qset), where);
+	  }
 
 	if (NILP (where))
 	  XSETBUFFER (where, current_buffer);
@@ -1790,6 +1802,14 @@ set_internal (Lisp_Object symbol, Lisp_Object newval, Lisp_Object where,
 	    && notify
 	    && bindflag != SET_INTERNAL_THREAD_SWITCH)
 	  {
+	    /* 'makunbound' nukes globally with zero regard of type
+	       (See code below, i.e. voide) do not send 'where'.
+	       If this is a 'BUFFER_OBJFWD' send 'where'.  */
+	    if (NILP (where)
+		&& !voide
+		&& BUFFER_OBJFWDP (innercontents))
+	      XSETBUFFER (where, current_buffer);
+
 	    notify_variable_watchers (symbol, voide? Qnil : newval,
 				      (bindflag == SET_INTERNAL_BIND? Qlet :
 				       bindflag == SET_INTERNAL_UNBIND? Qunlet :
@@ -1900,13 +1920,6 @@ notify_variable_watchers (Lisp_Object symbol,
   /* Avoid recursion.  */
   set_symbol_trapped_write (symbol, SYMBOL_UNTRAPPED_WRITE);
 
-  if (NILP (where)
-      && !EQ (operation, Qset_default) && !EQ (operation, Qmakunbound)
-      && !NILP (Flocal_variable_if_set_p (symbol, Fcurrent_buffer ())))
-    {
-      XSETBUFFER (where, current_buffer);
-    }
-
   if (EQ (operation, Qset_default))
     operation = Qset;
 
@@ -2038,7 +2051,16 @@ set_default_internal (Lisp_Object symbol, Lisp_Object value,
 	/* If watched notify unless thread switching */
 	if (sym->u.s.trapped_write == SYMBOL_TRAPPED_WRITE
 	    && bindflag != SET_INTERNAL_THREAD_SWITCH)
-	  notify_variable_watchers (symbol, value, Qset_default, Qnil);
+	  {
+	    /* Did we come from SPECPDL_LET_DEFAULT in do_one_unbind?
+	       Yes, then this let shadows default or make_local_foo
+	       was used for first time in this let, unlet is to default.
+	       If not from do_one_unbind this is a direct call.  */
+	    if (bindflag == SET_INTERNAL_UNBIND)
+	      notify_variable_watchers (symbol, value, Qunlet, Qnil);
+	    else
+	      notify_variable_watchers (symbol, value, Qset_default, Qnil);
+	  }
 
 	/* Store new value into the DEFAULT-VALUE slot.  */
 	XSETCDR (blv->defcell, value);
-- 
2.34.1

